%% -*- coding: utf-8 -*-
-module(ksudoku).
-author('hanak@inf.bme.hu, kapolnai@iit.bme.hu, modified by: Andras Eisenberger, Balazs Galambosi').
-vsn('$LastChangedDate: 2020-11-01 12:32:25 +0100 (v, 01 nov 2020) $$').

% -compile(export_all).
-export([sudoku_be/1,sudoku_ki/2,megold/2,stopper/2,stopper/3, help/0,teljes_teszt/1,teszt/2]).

-import(sudoku,[sudoku/1]).

-type sspec() :: {dist(), board()}.
-type dist()  :: non_neg_integer().
-type field() :: [info()].
-type info()  :: s | w | non_neg_integer().
-type board() :: [[field()]].

-type ssol() :: [[non_neg_integer()]].

%% -spec sudoku:sudoku(SSpec::sspec()) -> SSols::[ssol()].
%% SSols az SSpec feladványt kielégítő megoldások listája.

-spec sudoku_be(FileIn::string()) -> SSpec::sspec().
%% SSpec a FileIn szövegfájlból beolvasott Sudoku-feladvány.
sudoku_be(FileIn) ->
    case file:open(FileIn, [read]) of
	{ok, InputFile} ->
	    {N,_} = case io:get_line(InputFile,'') of
			eof ->
			    io:format("Üres a(z) ~s fájl.~n",[FileIn]),
			    exit('Hibás feladvány');
			I ->
			    string:to_integer(cleaned(I))
		    end,
	    T = tabla_be(InputFile),
	    file:close(InputFile),
%	    io:fwrite("~w\n",[T]),
	    KK = length(T),
	    if N < 1 orelse KK < N ->
		    io:format("a távolságkonstans kisebb 1-nél vagy nagyobb a sorhossznál.~n",[]),
		    exit('Hibás feladvány');
	       true ->
		    case helyes_feladvany(KK,T) of
			true ->
			    {N,T};
			false ->
			    io:format("Hibás a feladvány a(z) ~s fájlban.~n",[FileIn]),
			    exit('Hibás feladvány')
		    end
	    end
      ; _ ->
	    io:format("Nincs ~s nevű fájl.~n",[FileIn]),
	    exit('Nincs ilyen feladvány')
    end.

-spec tabla_be(identifier()) -> board().
tabla_be(InputFile) ->
    tabla_be(InputFile,[]).

-spec tabla_be(identifier(), board()) -> board().
tabla_be(InputFile,Zss) ->
    case io:get_line(InputFile,'') of
	eof ->
	    KK = length(Zss),
	    if KK < 1 orelse 25 < KK ->
		    io:format("1-nél kisebb vagy 5-nél nagyobb a cellaméret.~n",[]),
		    exit('Hibás feladvány');
	       true ->
		    case lists:all(fun(L) -> length(L) == KK end, Zss) of
			false ->
			    io:format("Nem négyzetes a feladvány, vagy nem egyforma hosszúak a sorok.~n",[]),
			    exit('Hibás feladvány');
			true ->
			    lists:reverse(Zss)
		    end
	    end
      ; Sor ->
	    Ss = string:tokens(cleaned(Sor)," "),
	    Zs = lists:map(fun str2infolist/1, Ss),
	    case Zs of
		[] ->
		    tabla_be(InputFile,Zss)
	      ; _ ->
		    tabla_be(InputFile,[Zs|Zss])
	    end
    end.

-spec cleaned(string()) -> string().
cleaned(Sor) ->
    S0 = lists:filter(fun(C) -> C /= $\t end,Sor),
    string:strip(string:strip(S0),right,$\n).

-spec str2infolist([char()]) -> [info()].
str2infolist([$-]) ->
    [];
str2infolist(XXs=[X|_Xs]) when $0=<X andalso X=<$9 ->
    {I,Rs} = string:to_integer(XXs),
    [I|str2infolist(Rs)];
str2infolist([C|Xs]) ->
    [list_to_atom([C])|str2infolist(Xs)];
str2infolist([]) ->
    [].

-spec helyes_feladvany(Size::non_neg_integer(),Board::board()) -> B::boolean().
%% B igaz, ha a feladvány megfelel a specifikációnak.
helyes_feladvany(KK,T) ->
    lists:all(fun(Fs) -> helyes_mezo(KK,Fs) end, lists:concat(T)).

-spec helyes_mezo(dist(),[info()]) -> boolean().
helyes_mezo(KK,Fs) ->
    helyes_mezo(KK,Fs,0,0,0).

-spec helyes_mezo(dist(),[info()],dist(),dist(),dist()) -> boolean().
helyes_mezo(_KK,[],S,V,W) ->
    S =< 1 andalso V =< 1 andalso W =< 1;
helyes_mezo(KK,[s|Fs],S,V,W) ->
    helyes_mezo(KK,Fs,S+1,V,W);
helyes_mezo(KK,[w|Fs],S,V,W) ->
    helyes_mezo(KK,Fs,S,V,W+1);
helyes_mezo(KK,[I|Fs],S,V,W) when is_integer(I) ->
    1 =< I andalso I =< KK andalso helyes_mezo(KK,Fs,S,V+1,W);
helyes_mezo(_KK,[_|_Fs],_S,_V,_W) ->
    io:format("Hibás infó van az input fájlban.~n",[]),
    exit('Hibás feladvány').

-spec tablak_ki(identifier(),[ssol()]) -> 'ok'.
tablak_ki(OutputFile,[SSol]) ->
    io:fwrite(OutputFile,"~s~n",[tabla_ki(SSol)]);
tablak_ki(OutputFile,[SSol|SSols]) ->
    io:fwrite(OutputFile,"~s~n~n",[tabla_ki(SSol)]),
    tablak_ki(OutputFile,SSols);
tablak_ki(OutputFile,[]) ->
    io:fwrite(OutputFile,"",[]).

-spec tabla_ki([string()]) -> [string()].
tabla_ki([L]) ->
    sor_ki(L);
tabla_ki([L|Ls]) ->
    sor_ki(L) ++ "\n" ++ tabla_ki(Ls);
tabla_ki(_) ->
    [].

-spec sor_ki([string()]) -> [string()].
sor_ki([H|T]) ->
    I2L = fun(Int) -> io_lib:format(if length([H|T]) > 9 -> "~2w" ; true -> "~w" end, [Int]) end,
    I2L(H) ++ lists:foldr(fun(A,B) -> " " ++ I2L(A) ++ B end,"",T);
sor_ki([]) ->
    "".

-spec sudoku_ki(FileOut::string(),SSols::[ssol()]) -> 'ok' | {'error',atom()}.
sudoku_ki(FileOut,SSols) ->
    {ok,OutputFile} = file:open(FileOut,[write]),
    io:setopts(OutputFile,[{encoding, unicode}]),
    tablak_ki(OutputFile,lists:sort(SSols)),
    file:close(OutputFile).

-spec sudokusol_be(FileIn::string()) -> SSols::[ssol()].
%% SSpec a FileIn szövegfájlból beolvasott Sudoku-feladvány.
%% Készítette: Eisenberger András.
sudokusol_be(FileSol) ->
    case file:open(FileSol, [read]) of
        {ok, InputFile}      -> ok
      ; {error, InputFile=R} -> throw({error,R,file:format_error(R)})
    end,
    case io:read(InputFile, "") of
        {ok, Sol}       -> ok
      ; {error, Sol=R2} -> throw({error,R2,file:format_error(R2)})
    end,
    file:close(InputFile),
    Sol.

-spec megold(FileIn::string(),FileOut::string()) -> no_return().
%% Beolvas egy feladványt a FileIn szövegfájlból, és összes megoldását
%% kiírja a FileOut szövegfájlba. Ehhez felhasználja a sudoku:sudoku/1
%% függvényt.
megold(FileIn,FileOut) ->
    KT = sudoku_be(FileIn),
    SSols = sudoku:sudoku(KT), 
    sudoku_ki(FileOut,SSols).

-spec stopper(FileIn::string(),FileOut::string()) -> no_return().
%% Mint megold/2, de a végén kiírja a FileIn nevet,
%% a megoldások számát és a futási időt is.
stopper(FileIn,FileOut) -> stopper(FileIn,FileOut,"").

-spec stopper(FileIn::string(),FileOut::string(),FileSol::string()) -> no_return().
%% Mint megold/2, de a végén kiírja a FileIn nevet, a megoldások
%% számát és a futási időt is. Ha lehet, összeveti a FileSol megoldással is.
stopper(FileIn,FileOut,FileSol) ->
    KT = sudoku_be(FileIn),
    erlang:garbage_collect(),timer:sleep(10),
    _T1 = statistics(runtime),
    SSols = sudoku:sudoku(KT),
    {_,T2} = statistics(runtime),
    sudoku_ki(FileOut,SSols),
    io:setopts([{encoding, unicode}]),
    io:format("Feladvány: ~s, ", [FileIn]),
    try   % Eisenberger András kiegészítése
        case lists:sort(sudokusol_be(FileSol)) =:= lists:sort(SSols) of
            true  -> io:format("HELYES ")
	  ; false -> io:format("ROSSZ ")
        end
    catch
        {error, _, _} -> error%io:format("NINCS MO ")
    end,
    io:format("megoldás: ~3w, futási idő: ",
	      [length(SSols)]),
    if T2 >= 10000 -> io:format("~5w s\n", [ T2 div 1000 ]);
%       T2 >= 1000 -> io:format("~5.2f s\n", [ T2 / 1000 ]);
       true        -> io:format("~5w ms\n", [ T2 ])
    end.

help() ->
    io:setopts([{encoding, unicode}]),
    io:format(
      "teljes_teszt(Timeout):~n"
      "  Megoldja az összes ./tests/testXXXd.txt feladványt a Timeout időlimittel.~n"
      "  A megoldásokat a ./tests_out_er/testXXXt.txt fájlokba rakja.~n"
      "  Példa: futtatás bash shell alatt, 10s időlimittel:~n"
      "    $ erl -noshell -eval 'ksudoku:teljes_teszt(10).' -s init stop~n"
     ).

-spec teljes_teszt(Timeout::integer()) -> [done|timeout].
%% A 'tests' könyvtárban levő összes "testXXXd.txt" tesztállomány esetén
%%  - lefuttatja a tesztet Timeout másodperces időkorláttal;
%%  - ellenőrzi, hogy a testXXXs.txt állományban megadott megoldáshalmazt kapta-e;
%%  - olvasható formában (lásd megold/2) kiírja az eredményt a 'tests_out_er'
%%    könyvtár testXXXt.txt nevű állományába.
%% Az állománynevekben az XXX szám tetszőleges hosszúságú lehet.
teljes_teszt(Timeout) ->
    Dir = "tests/",
    ODir = "tests_out_er/",
    case file:make_dir(ODir) of
        ok              -> ok
      ; {error, eexist} -> ok
      ; {error, Rd}     -> throw({error,Rd,file:format_error(Rd)})
    end,
    case file:list_dir(Dir) of
        {error, R} ->
            throw({error,R,file:format_error(R)})
      ; {ok, Content} ->
            IOFiles = [{Dir++Input,ODir++Output,Dir++Solution}
                       || Input <- lists:sort(Content),
                          {match,_} <- [re:run(Input, "test[0-9]*d\.txt")],
                          Solution <- [re:replace(Input, "d\.txt\$", "s.txt", [{return, list}])],
                          Output <- [re:replace(Input, "d\.txt\$", "t.txt", [{return, list}])]
                      ],
            teszt(Timeout, IOFiles)
    end.

-spec teszt(Timeout::integer(),
            IOFiles::[{Input::string(),Output::string(),Solution::string()}]) -> [done|timeout].
%% Megoldja az összes Input fájlban található specifikációt Timeout időlimittel, a
%% megoldásokat az Output fájlokba teszi. Ha lehet, összeveti a megoldásokat Solution-nal.
teszt(Timeout, IOFiles) ->
    [begin
         process_flag(trap_exit, true),
         Worker = spawn_link(fun() -> stopper(Input, Output, Solution) end),
         Timer  = spawn_link(fun() -> timer:sleep(Timeout * 1000) end),
         receive
             {'EXIT', Worker, _Reason} -> 
                 exit(Timer, kill),
                 receive {'EXIT', Timer, killed} -> ok end, % feldolgozzuk a halálát
                 done;
             {'EXIT', Timer, _Reason} ->
                 exit(Worker, kill),
                 io:format("Feladvány: ~s -- Timeout (~p s)~n", [Input,Timeout]),
                 receive {'EXIT', Worker, killed} -> ok end, % feldolgozzuk a halálát
                 timeout
         end
     end
     || {Input, Output, Solution} <- IOFiles
    ].
